/*
 * Copyright (C) 2023-2025 Xeme Developers
 *
 * Xeme is the legal property of its developers, whose names are too
 * numerous to list here. Please refer to the AUTHORS file distributed
 * with this source distribution
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, see
 * <https://www.gnu.org/licenses/>.
 */

#include "xemeaddress.h"

#include "xemestring.h"

#define XEME_ADDRESS_REGEX \
	"^(?P<bare_jid>" \
		"(?:(?P<local>[^\\\"&'/:<>@]+)@)?" \
		"(?P<domain>(?:[^@/[\\\\_$]+|(\\[[0-9a-fA-F:]+\\])))" \
	")" \
	"(?:/(?P<resource>.+))?$"

/******************************************************************************
 * Public API
 *****************************************************************************/
gboolean
xeme_address_parse(const char *address, char **bare_jid, char **local,
                   char **domain, char **resource)
{
	static GRegex *regexp = NULL;
	GMatchInfo *info = NULL;
	char *tmp = NULL;

	g_return_val_if_fail(!xeme_str_is_empty(address), FALSE);

	if(regexp == NULL) {
		GError *error = NULL;

		regexp = g_regex_new(XEME_ADDRESS_REGEX, G_REGEX_DEFAULT,
		                     G_REGEX_MATCH_DEFAULT, &error);

		if(error != NULL) {
			g_critical("failed to compiled the regular expression for "
			           "addresses %s",
			           error->message != NULL ? error->message : "unknown");

			g_clear_error(&error);
		}
	}

	if(!g_regex_match(regexp, address, G_REGEX_MATCH_DEFAULT, &info)) {
		g_clear_pointer(&info, g_match_info_free);

		return FALSE;
	}

	if(bare_jid != NULL) {
		tmp = g_match_info_fetch_named(info, "bare_jid");
		if(xeme_str_is_empty(tmp)) {
			g_clear_pointer(&tmp, g_free);
		}

		*bare_jid = tmp;
	}

	if(local != NULL) {
		tmp = g_match_info_fetch_named(info, "local");
		if(xeme_str_is_empty(tmp)) {
			g_clear_pointer(&tmp, g_free);
		}

		*local = tmp;
	}

	if(domain != NULL) {
		tmp = g_match_info_fetch_named(info, "domain");
		if(xeme_str_is_empty(tmp)) {
			g_clear_pointer(&tmp, g_free);
		}

		*domain = tmp;
	}

	if(resource != NULL) {
		tmp = g_match_info_fetch_named(info, "resource");
		if(xeme_str_is_empty(tmp)) {
			g_clear_pointer(&tmp, g_free);
		}

		*resource = tmp;
	}

	g_clear_pointer(&info, g_match_info_free);

	return TRUE;
}
