/*
 * Copyright (C) 2023-2025 Xeme Developers
 *
 * Xeme is the legal property of its developers, whose names are too
 * numerous to list here. Please refer to the AUTHORS file distributed
 * with this source distribution
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, see
 * <https://www.gnu.org/licenses/>.
 */

#include "xemeconnection.h"

#include "xemeaddress.h"
#include "xemeconstants.h"
#include "xemeenums.h"
#include "xemestring.h"

enum {
	PROP_0,
	PROP_JID,
	PROP_RESOURCE,
	PROP_SERVER,
	PROP_PORT,
	PROP_TLS_MODE,
	N_PROPERTIES,
};
static GParamSpec *properties[N_PROPERTIES] = {NULL, };

struct _XemeConnection {
	GObject parent;

	char *jid;
	char *domain; /* This is parsed out of the jid when it is set. */
	char *resource;

	char *server;
	guint16 port;

	XemeTlsMode tls_mode;

	GSocketClient *client;
	GSocketConnection *socket_connection;

	XemeInputStream *input;

	GCancellable *cancellable;
};

G_DEFINE_FINAL_TYPE(XemeConnection, xeme_connection, G_TYPE_OBJECT)

/******************************************************************************
 * Helpers
 *****************************************************************************/
static void
xeme_connection_set_jid(XemeConnection *connection, const char *jid) {
	char *bare_jid = NULL;
	char *domain = NULL;
	char *resource = NULL;

	if(!xeme_address_parse(jid, &bare_jid, NULL, &domain, &resource)) {
		g_critical("invalid jid '%s'", jid);

		return;
	}

	g_free(connection->jid);
	connection->jid = bare_jid;
	connection->domain = domain;

	if(resource != NULL) {
		xeme_connection_set_resource(connection, resource);
		g_free(resource);
	}
}

static void
xeme_connection_connect_success(XemeConnection *connection, GTask *task) {
	XemeOutputStream *output = NULL;
	GError *error = NULL;
	GIOStream *iostream = NULL;
	GInputStream *istream = NULL;
	GOutputStream *ostream = NULL;

	iostream = G_IO_STREAM(connection->socket_connection);

	ostream = g_io_stream_get_output_stream(iostream);
	output = xeme_output_stream_new(connection->cancellable);
	xeme_stream_set_to(XEME_STREAM(output), connection->domain);
	xeme_stream_set_from(XEME_STREAM(output), connection->jid);
	if(!xeme_output_stream_start(output, ostream, &error)) {
		g_task_return_error(task, error);

		g_clear_object(&task);
		g_clear_object(&output);

		xeme_connection_close(connection, NULL);

		return;
	}

	istream = g_io_stream_get_input_stream(iostream);
	if(!xeme_input_stream_start(connection->input, istream, output, &error)) {
		g_task_return_error(task, error);

		g_clear_object(&task);
		g_clear_object(&output);

		xeme_connection_close(connection, NULL);

		return;
	}

	/* Finally return the output stream. */
	g_task_return_pointer(task, output, g_object_unref);

	g_clear_object(&task);
}

/******************************************************************************
 * connect_to_host functions
 *****************************************************************************/
static void
xeme_connection_connect_to_host_cb(GObject *source, GAsyncResult *result,
                                   gpointer data)
{
	XemeConnection *connection = NULL;
	GError *error = NULL;
	GSocketClient *client = G_SOCKET_CLIENT(source);
	GSocketConnection *socket_connection = NULL;
	GTask *task = data;

	connection = g_task_get_source_object(task);
	socket_connection = g_socket_client_connect_to_host_finish(client, result,
	                                                           &error);
	if(error != NULL) {
		g_task_return_error(task, error);
		g_clear_object(&task);

		return;
	}

	connection->socket_connection = socket_connection;

	xeme_connection_connect_success(connection, task);
}

static void
xeme_connection_connect_to_host(XemeConnection *connection, GTask *task) {
	gboolean use_tls = FALSE;
	const char *server = NULL;

	g_return_if_fail(XEME_IS_CONNECTION(connection));
	g_return_if_fail(G_IS_TASK(task));

	use_tls = (connection->tls_mode == XEME_TLS_MODE_DIRECT_TLS);
	g_socket_client_set_tls(connection->client, use_tls);

	server = connection->server;
	if(xeme_str_is_empty(server)) {
		server = connection->domain;
	}

	g_socket_client_connect_to_host_async(connection->client,
	                                      server,
	                                      connection->port,
	                                      connection->cancellable,
	                                      xeme_connection_connect_to_host_cb,
	                                      task);
}

/******************************************************************************
 * connect_to_service functions
 *****************************************************************************/
static void
xeme_connection_connect_to_service_cb(GObject *source, GAsyncResult *result,
                                      gpointer data)
{
	XemeConnection *connection = NULL;
	GError *error = NULL;
	GSocketClient *client = G_SOCKET_CLIENT(source);
	GSocketConnection *socket_connection = NULL;
	GTask *task = data;

	connection = g_task_get_source_object(task);
	socket_connection = g_socket_client_connect_to_service_finish(client,
	                                                              result,
	                                                              &error);
	if(error != NULL) {
		g_warning("failed to connect via srv record, attempting direct "
		          "connect.");

		xeme_connection_connect_to_host(connection, task);

		return;
	}

	connection->socket_connection = socket_connection;

	xeme_connection_connect_success(connection, task);
}

static void
xeme_connection_connect_to_service(XemeConnection *connection, GTask *task) {
	gboolean use_tls = TRUE;
	const char *service = XEME_SERVICE_XMPPS_CLIENT;

	g_return_if_fail(XEME_IS_CONNECTION(connection));
	g_return_if_fail(G_IS_TASK(task));

	if(connection->tls_mode != XEME_TLS_MODE_DIRECT_TLS) {
		service = XEME_SERVICE_XMPP_CLIENT;
		use_tls = FALSE;
	}

	g_socket_client_set_tls(connection->client, use_tls);

	g_socket_client_connect_to_service_async(connection->client,
	                                         connection->domain,
	                                         service,
	                                         connection->cancellable,
	                                         xeme_connection_connect_to_service_cb,
	                                         task);
}

/******************************************************************************
 * GObject Implementation
 *****************************************************************************/
static void
xeme_connection_finalize(GObject *obj) {
	XemeConnection *connection = XEME_CONNECTION(obj);

	xeme_connection_close(connection, NULL);

	g_clear_pointer(&connection->jid, g_free);
	g_clear_pointer(&connection->domain, g_free);
	g_clear_pointer(&connection->resource, g_free);
	g_clear_pointer(&connection->server, g_free);

	g_clear_object(&connection->client);
	g_clear_object(&connection->socket_connection);

	g_clear_object(&connection->input);

	g_clear_object(&connection->cancellable);

	G_OBJECT_CLASS(xeme_connection_parent_class)->finalize(obj);
}

static void
xeme_connection_get_property(GObject *obj, guint param_id, GValue *value,
                             GParamSpec *pspec)
{
	XemeConnection *connection = XEME_CONNECTION(obj);

	switch(param_id) {
	case PROP_JID:
		g_value_set_string(value, xeme_connection_get_jid(connection));
		break;
	case PROP_RESOURCE:
		g_value_set_string(value, xeme_connection_get_resource(connection));
		break;
	case PROP_SERVER:
		g_value_set_string(value, xeme_connection_get_server(connection));
		break;
	case PROP_PORT:
		g_value_set_uint(value, xeme_connection_get_port(connection));
		break;
	case PROP_TLS_MODE:
		g_value_set_enum(value, xeme_connection_get_tls_mode(connection));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_connection_set_property(GObject *obj, guint param_id, const GValue *value,
                             GParamSpec *pspec)
{
	XemeConnection *connection = XEME_CONNECTION(obj);

	switch(param_id) {
	case PROP_JID:
		xeme_connection_set_jid(connection, g_value_get_string(value));
		break;
	case PROP_RESOURCE:
		xeme_connection_set_resource(connection, g_value_get_string(value));
		break;
	case PROP_SERVER:
		xeme_connection_set_server(connection, g_value_get_string(value));
		break;
	case PROP_PORT:
		xeme_connection_set_port(connection, g_value_get_uint(value));
		break;
	case PROP_TLS_MODE:
		xeme_connection_set_tls_mode(connection, g_value_get_enum(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_connection_init(G_GNUC_UNUSED XemeConnection *connection) {
}

static void
xeme_connection_class_init(XemeConnectionClass *klass) {
	GObjectClass *obj_class = G_OBJECT_CLASS(klass);

	obj_class->finalize = xeme_connection_finalize;
	obj_class->get_property = xeme_connection_get_property;
	obj_class->set_property = xeme_connection_set_property;

	/**
	 * XemeConnection:jid:
	 *
	 * The "Jabber ID" which this connection originates from.
	 *
	 * Since: 0.1
	 */
	properties[PROP_JID] = g_param_spec_string(
		"jid", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeConnection:resource:
	 *
	 * The "resource" for the originator of this connection.
	 *
	 * If the `jid` passed to [ctor@Connection.new] includes a resource, it
	 * will be parsed out and saved here.
	 *
	 * If this is %NULL during connection, @connection will request one from
	 * the server and store it here. The server provided resource should be
	 * saved by the application and set by the application in the future.
	 *
	 * Resources are mostly deprecated, but this property can be modify to set
	 * it to whatever you like.
	 *
	 * Since: 0.1
	 */
	properties[PROP_RESOURCE] = g_param_spec_string(
		"resource", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeConnection:server:
	 *
	 * An explicit server to connect to.
	 *
	 * If this is set, it will be the only connection attempt made. See
	 * [method@Connection.connect_async] for more information.
	 *
	 * Since: 0.1
	 */
	properties[PROP_SERVER] = g_param_spec_string(
		"server", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeConnection:port:
	 *
	 * The explicit port to use.
	 *
	 * This will only be used if [property@Connection:server] is not %NULL.
	 *
	 * Since: 0.1
	 */
	properties[PROP_PORT] = g_param_spec_uint(
		"port", NULL, NULL,
		0, G_MAXUINT16, XEME_PORT_TLS,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeConnection:tls-mode:
	 *
	 * The TLS mode to use for this connection.
	 *
	 * This determines whether or not a TLS connection is attempted at all or
	 * if STARTTLS is used to upgrade a clear connection.
	 *
	 * Since: 0.1
	 */
	properties[PROP_TLS_MODE] = g_param_spec_enum(
		"tls-mode", NULL, NULL,
		XEME_TYPE_TLS_MODE,
		XEME_TLS_MODE_DIRECT_TLS,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	g_object_class_install_properties(obj_class, N_PROPERTIES, properties);
}

/******************************************************************************
 * Public API
 *****************************************************************************/
XemeConnection *
xeme_connection_new(const char *jid) {
	g_return_val_if_fail(!xeme_str_is_empty(jid), NULL);

	return g_object_new(
		XEME_TYPE_CONNECTION,
		"jid", jid,
		NULL);
}

const char *
xeme_connection_get_jid(XemeConnection *connection) {
	g_return_val_if_fail(XEME_IS_CONNECTION(connection), NULL);

	return connection->jid;
}

const char *
xeme_connection_get_resource(XemeConnection *connection) {
	g_return_val_if_fail(XEME_IS_CONNECTION(connection), NULL);

	return connection->resource;
}

void
xeme_connection_set_resource(XemeConnection *connection, const char *resource)
{
	g_return_if_fail(XEME_IS_CONNECTION(connection));

	if(g_set_str(&connection->resource, resource)) {
		g_object_notify_by_pspec(G_OBJECT(connection),
		                         properties[PROP_RESOURCE]);
	}
}

const char *
xeme_connection_get_server(XemeConnection *connection) {
	g_return_val_if_fail(XEME_IS_CONNECTION(connection), NULL);

	return connection->server;
}

void
xeme_connection_set_server(XemeConnection *connection, const char *server) {
	g_return_if_fail(XEME_IS_CONNECTION(connection));

	if(g_set_str(&connection->server, server)) {
		g_object_notify_by_pspec(G_OBJECT(connection),
		                         properties[PROP_SERVER]);
	}
}

guint16
xeme_connection_get_port(XemeConnection *connection) {
	g_return_val_if_fail(XEME_IS_CONNECTION(connection), 0);

	return connection->port;
}

void
xeme_connection_set_port(XemeConnection *connection, guint16 port) {
	g_return_if_fail(XEME_IS_CONNECTION(connection));

	if(connection->port != port) {
		connection->port = port;

		g_object_notify_by_pspec(G_OBJECT(connection), properties[PROP_PORT]);
	}
}

XemeTlsMode
xeme_connection_get_tls_mode(XemeConnection *connection) {
	g_return_val_if_fail(XEME_IS_CONNECTION(connection), XEME_TLS_MODE_DIRECT_TLS);

	return connection->tls_mode;
}

void
xeme_connection_set_tls_mode(XemeConnection *connection, XemeTlsMode tls_mode)
{
	g_return_if_fail(XEME_IS_CONNECTION(connection));

	if(connection->tls_mode != tls_mode) {
		connection->tls_mode = tls_mode;

		g_object_notify_by_pspec(G_OBJECT(connection),
		                         properties[PROP_TLS_MODE]);
	}
}

void
xeme_connection_connect_async(XemeConnection *connection,
                              XemeInputStream *input,
                              GProxyResolver *resolver,
                              GCancellable *cancellable,
                              GAsyncReadyCallback callback,
                              gpointer data)
{
	GTask *task = NULL;

	g_return_if_fail(XEME_IS_CONNECTION(connection));
	g_return_if_fail(XEME_IS_INPUT_STREAM(input));

	if(G_IS_CANCELLABLE(cancellable)) {
		connection->cancellable = g_object_ref(cancellable);
	}

	connection->input = g_object_ref(input);

	connection->client = g_socket_client_new();
	if(G_IS_PROXY_RESOLVER(resolver)) {
		g_socket_client_set_proxy_resolver(connection->client, resolver);
		g_clear_object(&resolver);
	}

	task = g_task_new(connection, cancellable, callback, data);

	/* We're doing a direct connect to a server. */
	if(!xeme_str_is_empty(connection->server)) {
		xeme_connection_connect_to_host(connection, task);
	} else {
		xeme_connection_connect_to_service(connection, task);
	}
}

XemeOutputStream *
xeme_connection_connect_finish(XemeConnection *connection,
                               GAsyncResult *result,
                               GError **error)
{
	g_return_val_if_fail(XEME_IS_CONNECTION(connection), FALSE);
	g_return_val_if_fail(G_IS_ASYNC_RESULT(result), FALSE);

	return g_task_propagate_pointer(G_TASK(result), error);
}

gboolean
xeme_connection_close(XemeConnection *connection, GError **error) {
	g_return_val_if_fail(XEME_IS_CONNECTION(connection), FALSE);

	if(G_IS_IO_STREAM(connection->socket_connection)) {
		/* We're shutting down so we ignore errors. */
		g_io_stream_close(G_IO_STREAM(connection->socket_connection), NULL,
		                  error);
		g_clear_object(&connection->socket_connection);
	}

	g_clear_object(&connection->input);

	return TRUE;
}
