/*
 * Copyright (C) 2023-2025 Xeme Developers
 *
 * Xeme is the legal property of its developers, whose names are too
 * numerous to list here. Please refer to the AUTHORS file distributed
 * with this source distribution
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, see
 * <https://www.gnu.org/licenses/>.
 */

#include "xemeextension.h"

#include "xemestring.h"

enum {
	PROP_0,
	PROP_ELEMENT,
	PROP_NAMESPACE,
	N_PROPERTIES,
};
static GParamSpec *properties[N_PROPERTIES] = {NULL, };

typedef struct {
	char *element;
	char *namespace;
} XemeExtensionPrivate;

G_DEFINE_TYPE_WITH_PRIVATE(XemeExtension, xeme_extension, G_TYPE_OBJECT)

/******************************************************************************
 * Helpers
 *****************************************************************************/
static void
xeme_extension_set_element(XemeExtension *extension, const char *element) {
	XemeExtensionPrivate *priv = NULL;

	g_return_if_fail(XEME_IS_EXTENSION(extension));

	priv = xeme_extension_get_instance_private(extension);

	if(g_set_str(&priv->element, element)) {
		g_object_notify_by_pspec(G_OBJECT(extension),
		                         properties[PROP_ELEMENT]);
	}
}

static void
xeme_extension_set_namespace(XemeExtension *extension, const char *namespace) {
	XemeExtensionPrivate *priv = NULL;

	g_return_if_fail(XEME_IS_EXTENSION(extension));

	priv = xeme_extension_get_instance_private(extension);

	if(g_set_str(&priv->namespace, namespace)) {
		g_object_notify_by_pspec(G_OBJECT(extension),
		                         properties[PROP_NAMESPACE]);
	}
}

/******************************************************************************
 * GObject Implementation
 *****************************************************************************/
static void
xeme_extension_finalize(GObject *obj) {
	XemeExtension *extension = XEME_EXTENSION(obj);
	XemeExtensionPrivate *priv = NULL;

	priv = xeme_extension_get_instance_private(extension);

	g_clear_pointer(&priv->element, g_free);
	g_clear_pointer(&priv->namespace, g_free);

	G_OBJECT_CLASS(xeme_extension_parent_class)->finalize(obj);
}

static void
xeme_extension_get_property(GObject *obj, guint param_id, GValue *value,
                            GParamSpec *pspec)
{
	XemeExtension *extension = XEME_EXTENSION(obj);

	switch(param_id) {
	case PROP_ELEMENT:
		g_value_set_string(value, xeme_extension_get_element(extension));
		break;
	case PROP_NAMESPACE:
		g_value_set_string(value, xeme_extension_get_namespace(extension));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_extension_set_property(GObject *obj, guint param_id, const GValue *value,
                            GParamSpec *pspec)
{
	XemeExtension *extension = XEME_EXTENSION(obj);

	switch(param_id) {
	case PROP_ELEMENT:
		xeme_extension_set_element(extension, g_value_get_string(value));
		break;
	case PROP_NAMESPACE:
		xeme_extension_set_namespace(extension, g_value_get_string(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_extension_init(G_GNUC_UNUSED XemeExtension *extension) {
}

static void
xeme_extension_class_init(XemeExtensionClass *klass) {
	GObjectClass *obj_class = G_OBJECT_CLASS(klass);

	obj_class->finalize = xeme_extension_finalize;
	obj_class->get_property = xeme_extension_get_property;
	obj_class->set_property = xeme_extension_set_property;

	/**
	 * XemeExtension:element:
	 *
	 * The name of the element that's being extended.
	 *
	 * Since: 0.1
	 */
	properties[PROP_ELEMENT] = g_param_spec_string(
		"element", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeExtension:namespace:
	 *
	 * The namespace of the extension.
	 *
	 * Since: 0.1
	 */
	properties[PROP_NAMESPACE] = g_param_spec_string(
		"namespace", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);

	g_object_class_install_properties(obj_class, N_PROPERTIES, properties);
}

/******************************************************************************
 * Public API
 *****************************************************************************/
const char *
xeme_extension_get_element(XemeExtension *extension) {
	XemeExtensionPrivate *priv = NULL;

	g_return_val_if_fail(XEME_IS_EXTENSION(extension), NULL);

	priv = xeme_extension_get_instance_private(extension);

	return priv->element;
}

const char *
xeme_extension_get_namespace(XemeExtension *extension) {
	XemeExtensionPrivate *priv = NULL;

	g_return_val_if_fail(XEME_IS_EXTENSION(extension), NULL);

	priv = xeme_extension_get_instance_private(extension);

	return priv->namespace;
}

void
xeme_extension_parse_start(XemeExtension *extension, XemeInputStream *input,
                           GMarkupParseContext *context)
{
	XemeExtensionClass *klass = NULL;

	g_return_if_fail(XEME_IS_EXTENSION(extension));
	g_return_if_fail(XEME_IS_INPUT_STREAM(input));
	g_return_if_fail(context != NULL);

	klass = XEME_EXTENSION_GET_CLASS(extension);
	if(klass != NULL && klass->parse_start != NULL) {
		klass->parse_start(extension, input, context);
	}
}

void
xeme_extension_parse_end(XemeExtension *extension,
                         GMarkupParseContext *context)
{
	XemeExtensionClass *klass = NULL;

	g_return_if_fail(XEME_IS_EXTENSION(extension));
	g_return_if_fail(context != NULL);

	klass = XEME_EXTENSION_GET_CLASS(extension);
	if(klass != NULL && klass->parse_end != NULL) {
		klass->parse_end(extension, context);
	}
}

char *
xeme_extension_serialize(XemeExtension *extension) {
	XemeExtensionClass *klass = NULL;

	g_return_val_if_fail(XEME_IS_EXTENSION(extension), NULL);

	klass = XEME_EXTENSION_GET_CLASS(extension);
	if(klass != NULL && klass->serialize != NULL) {
		return klass->serialize(extension);
	}

	return NULL;
}
