/*
 * Copyright (C) 2023-2025 Xeme Developers
 *
 * Xeme is the legal property of its developers, whose names are too
 * numerous to list here. Please refer to the AUTHORS file distributed
 * with this source distribution
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, see
 * <https://www.gnu.org/licenses/>.
 */

#include "xemeextensionmap.h"

#include "xemestring.h"

#define XEME_EXTENSION_MAP_DELIMITER ("@")

typedef struct {
	XemeExtensionMap *map;
	XemeInputStream *input;
	XemeExtension *extension;
} XemeExtensionMapParseData;

struct _XemeExtensionMap {
	GObject parent;

	GHashTable *extensions;
};

G_DEFINE_FINAL_TYPE(XemeExtensionMap, xeme_extension_map, G_TYPE_OBJECT)

/******************************************************************************
 * Helpers
 *****************************************************************************/
static XemeExtensionMapParseData *
xeme_extension_map_parse_data_new(XemeExtensionMap *map,
                                  XemeInputStream *input)
{
	XemeExtensionMapParseData *data = NULL;

	data = g_new(XemeExtensionMapParseData, 1);
	data->map = g_object_ref(map);
	data->input = g_object_ref(input);
	data->extension = NULL;

	return data;
}

static void
xeme_extension_map_parse_data_free(XemeExtensionMapParseData *data) {
	g_return_if_fail(data != NULL);

	g_clear_object(&data->map);
	g_clear_object(&data->input);
	g_clear_object(&data->extension);

	g_clear_pointer(&data, g_free);
}

static inline char *
xeme_extension_map_generate_key(const char *element, const char *namespace) {
	return g_strconcat(element, XEME_EXTENSION_MAP_DELIMITER, namespace, NULL);
}

static inline char *
xeme_extension_map_generate_key_from_extension(XemeExtension *extension) {
	const char *element = NULL;
	const char *namespace = NULL;

	element = xeme_extension_get_element(extension);
	g_return_val_if_fail(!xeme_str_is_empty(element), NULL);

	namespace = xeme_extension_get_namespace(extension);
	g_return_val_if_fail(!xeme_str_is_empty(namespace), NULL);

	return xeme_extension_map_generate_key(element, namespace);
}

/******************************************************************************
 * Parser
 *****************************************************************************/
static void
xeme_extension_map_parser_element_start(GMarkupParseContext *context,
                                        const char *element_name,
                                        const char **attribute_names,
                                        const char **attribute_values,
                                        gpointer user_data,
                                        GError **error)
{
	XemeExtensionMapParseData *data = user_data;
	XemeExtension *extension = NULL;
	gboolean ret = FALSE;
	const char *namespace = NULL;

	ret = g_markup_collect_attributes(
		element_name, attribute_names, attribute_values, error,
		G_MARKUP_COLLECT_STRING, "xmlns", &namespace,
		G_MARKUP_COLLECT_INVALID);

	if(!ret) {
		return;
	}

	extension = xeme_extension_map_lookup(data->map, element_name, namespace);
	if(!XEME_IS_EXTENSION(extension)) {
		g_message("unknown extension element '%s' with namespace '%s'",
		          element_name, namespace);
		return;
	}

	g_set_object(&data->extension, extension);
	xeme_extension_parse_start(extension, data->input, context);
}

static void
xeme_extension_map_parser_element_end(GMarkupParseContext *context,
                                      G_GNUC_UNUSED const char *element_name,
                                      gpointer user_data,
                                      G_GNUC_UNUSED GError **error)
{
	XemeExtensionMapParseData *data = user_data;

	/* data->extension will be null for extensions we don't know about. */
	if(!XEME_IS_EXTENSION(data->extension)) {
		return;
	}

	xeme_extension_parse_end(data->extension, context);
	g_clear_object(&data->extension);
}

static GMarkupParser parser = {
	.start_element = xeme_extension_map_parser_element_start,
	.end_element = xeme_extension_map_parser_element_end,
};

/******************************************************************************
 * GObject Implementation
 *****************************************************************************/
static void
xeme_extension_map_finalize(GObject *obj) {
	XemeExtensionMap *map = XEME_EXTENSION_MAP(obj);

	g_clear_pointer(&map->extensions, g_hash_table_unref);

	G_OBJECT_CLASS(xeme_extension_map_parent_class)->finalize(obj);
}

static void
xeme_extension_map_init(XemeExtensionMap *map) {
	map->extensions = g_hash_table_new_full(g_str_hash, g_str_equal, g_free,
	                                        g_object_unref);
}

static void
xeme_extension_map_class_init(XemeExtensionMapClass *klass) {
	GObjectClass *obj_class = G_OBJECT_CLASS(klass);

	obj_class->finalize = xeme_extension_map_finalize;
}

/******************************************************************************
 * Public API
 *****************************************************************************/
XemeExtensionMap *
xeme_extension_map_new(void) {
	return g_object_new(XEME_TYPE_EXTENSION_MAP, NULL);
}

gboolean
xeme_extension_map_add(XemeExtensionMap *map, XemeExtension *extension) {
	char *key = NULL;

	g_return_val_if_fail(XEME_IS_EXTENSION_MAP(map), FALSE);
	g_return_val_if_fail(XEME_IS_EXTENSION(extension), FALSE);

	key = xeme_extension_map_generate_key_from_extension(extension);
	g_return_val_if_fail(key != NULL, FALSE);

	return g_hash_table_insert(map->extensions, key, g_object_ref(extension));
}

gboolean
xeme_extension_map_remove(XemeExtensionMap *map, XemeExtension *extension) {
	char *key = NULL;
	gboolean ret = FALSE;

	g_return_val_if_fail(XEME_IS_EXTENSION_MAP(map), FALSE);
	g_return_val_if_fail(XEME_IS_EXTENSION(extension), FALSE);

	key = xeme_extension_map_generate_key_from_extension(extension);
	g_return_val_if_fail(key != NULL, FALSE);

	ret = g_hash_table_remove(map->extensions, key);
	g_free(key);

	return ret;
}

XemeExtension *
xeme_extension_map_lookup(XemeExtensionMap *map, const char *element,
                          const char *namespace)
{
	XemeExtension *extension = NULL;
	char *key = NULL;

	g_return_val_if_fail(XEME_IS_EXTENSION_MAP(map), NULL);
	g_return_val_if_fail(!xeme_str_is_empty(element), NULL);
	g_return_val_if_fail(!xeme_str_is_empty(namespace), NULL);

	key = xeme_extension_map_generate_key(element, namespace);
	extension = g_hash_table_lookup(map->extensions, key);
	g_free(key);

	return extension;
}

guint
xeme_extension_map_get_size(XemeExtensionMap *map) {
	g_return_val_if_fail(XEME_IS_EXTENSION_MAP(map), 0);

	return g_hash_table_size(map->extensions);
}

void
xeme_extension_map_parse_start(XemeExtensionMap *map, XemeInputStream *input,
                               GMarkupParseContext *context)
{
	XemeExtensionMapParseData *data = NULL;

	g_return_if_fail(XEME_IS_EXTENSION_MAP(map));
	g_return_if_fail(XEME_IS_INPUT_STREAM(input));
	g_return_if_fail(context != NULL);

	data = xeme_extension_map_parse_data_new(map, input);

	g_markup_parse_context_push(context, &parser, data);
}

void
xeme_extension_map_parse_end(XemeExtensionMap *map,
                             GMarkupParseContext *context)
{
	XemeExtensionMapParseData *data = NULL;

	g_return_if_fail(XEME_IS_EXTENSION_MAP(map));
	g_return_if_fail(context != NULL);

	data = g_markup_parse_context_pop(context);

	xeme_extension_map_parse_data_free(data);
}
