/*
 * Copyright (C) 2023-2025 Xeme Developers
 *
 * Xeme is the legal property of its developers, whose names are too
 * numerous to list here. Please refer to the AUTHORS file distributed
 * with this source distribution
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, see
 * <https://www.gnu.org/licenses/>.
 */

#include <birb.h>

#include "xemelocalizedstring.h"

#include "xemestring.h"

struct _XemeLocalizedString {
	GObject parent;

	char *lang;
	char *str;
};

enum {
	PROP_0,
	PROP_LANGUAGE,
	PROP_STRING,
	N_PROPERTIES,
};
static GParamSpec *properties[N_PROPERTIES] = {NULL, 0};

/******************************************************************************
 * GObject Implementation
 *****************************************************************************/
G_DEFINE_FINAL_TYPE(XemeLocalizedString, xeme_localized_string, G_TYPE_OBJECT)

static void
xeme_localized_string_finalize(GObject *obj) {
	XemeLocalizedString *l10nstr = XEME_LOCALIZED_STRING(obj);

	g_clear_pointer(&l10nstr->lang, g_free);
	g_clear_pointer(&l10nstr->str, g_free);

	G_OBJECT_CLASS(xeme_localized_string_parent_class)->finalize(obj);
}

static void
xeme_localized_string_get_property(GObject *obj, guint param_id, GValue *value,
                                   GParamSpec *pspec)
{
	XemeLocalizedString *l10nstr = XEME_LOCALIZED_STRING(obj);

	switch(param_id) {
	case PROP_LANGUAGE:
		g_value_set_string(value, xeme_localized_string_get_language(l10nstr));
		break;
	case PROP_STRING:
		g_value_set_string(value, xeme_localized_string_get_string(l10nstr));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_localized_string_set_property(GObject *obj, guint param_id,
                                   const GValue *value, GParamSpec *pspec)
{
	XemeLocalizedString *l10nstr = XEME_LOCALIZED_STRING(obj);

	switch(param_id) {
	case PROP_LANGUAGE:
		xeme_localized_string_set_language(l10nstr, g_value_get_string(value));
		break;
	case PROP_STRING:
		xeme_localized_string_set_string(l10nstr, g_value_get_string(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_localized_string_init(G_GNUC_UNUSED XemeLocalizedString *l10nstr) {
}

static void
xeme_localized_string_class_init(XemeLocalizedStringClass *klass) {
	GObjectClass *obj_class = G_OBJECT_CLASS(klass);

	obj_class->finalize = xeme_localized_string_finalize;
	obj_class->get_property = xeme_localized_string_get_property;
	obj_class->set_property = xeme_localized_string_set_property;

	/**
	 * XemeLocalizedString:language:
	 *
	 * The language for this string.
	 *
	 * This may be %NULL.
	 *
	 * Since: 0.1
	 */
	properties[PROP_LANGUAGE] = g_param_spec_string(
		"language", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeLocalizedString:string:
	 *
	 * The string.
	 *
	 * Since: 0.1
	 */
	properties[PROP_STRING] = g_param_spec_string(
		"string", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	g_object_class_install_properties(obj_class, N_PROPERTIES, properties);
}

/******************************************************************************
 * Public API
 *****************************************************************************/
gboolean
xeme_localized_string_equal(XemeLocalizedString *l10nstr1,
                            XemeLocalizedString *l10nstr2)
{
	if(l10nstr1 == l10nstr2) {
		return TRUE;
	}

	if(l10nstr1 == NULL || l10nstr2 == NULL) {
		return FALSE;
	}

	if(xeme_str_equal(l10nstr1->lang, l10nstr2->lang)) {
		if(xeme_str_equal(l10nstr1->str, l10nstr2->str)) {
			return TRUE;
		}
	}

	return FALSE;
}

const char *
xeme_localized_string_get_language(XemeLocalizedString *l10nstr) {
	g_return_val_if_fail(XEME_IS_LOCALIZED_STRING(l10nstr), NULL);

	return l10nstr->lang;
}

const char *
xeme_localized_string_get_string(XemeLocalizedString *l10nstr) {
	g_return_val_if_fail(XEME_IS_LOCALIZED_STRING(l10nstr), NULL);

	return l10nstr->str;
}

XemeLocalizedString *
xeme_localized_string_new(const char *lang, const char *str) {
	g_return_val_if_fail(!xeme_str_is_empty(str), NULL);

	return g_object_new(
		XEME_TYPE_LOCALIZED_STRING,
		"language", lang,
		"string", str,
		NULL);
}

void
xeme_localized_string_set_language(XemeLocalizedString *l10nstr,
                                   const char *lang)
{
	g_return_if_fail(XEME_IS_LOCALIZED_STRING(l10nstr));

	if(g_set_str(&l10nstr->lang, lang)) {
		g_object_notify_by_pspec(G_OBJECT(l10nstr), properties[PROP_LANGUAGE]);
	}
}

void
xeme_localized_string_set_string(XemeLocalizedString *l10nstr, const char *str)
{
	g_return_if_fail(XEME_IS_LOCALIZED_STRING(l10nstr));
	g_return_if_fail(!xeme_str_is_empty(str));

	if(g_set_str(&l10nstr->str, str)) {
		g_object_notify_by_pspec(G_OBJECT(l10nstr), properties[PROP_STRING]);
	}
}
