/*
 * Copyright (C) 2023-2025 Xeme Developers
 *
 * Xeme is the legal property of its developers, whose names are too
 * numerous to list here. Please refer to the AUTHORS file distributed
 * with this source distribution
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, see
 * <https://www.gnu.org/licenses/>.
 */

#include "xememessage.h"

#include "xemeenums.h"
#include "xemestring.h"

enum {
	PROP_0,
	PROP_BODIES,
	PROP_MESSAGE_TYPE,
	PROP_SUBJECTS,
	PROP_THREAD,
	PROP_THREAD_PARENT,
	N_PROPERTIES,
};
static GParamSpec *properties[N_PROPERTIES] = {NULL, };

struct _XemeMessage {
	XemeStanza parent;

	XemeMessageType message_type;

	GListStore *bodies;
	GListStore *subjects;

	char *thread;
	char *thread_parent;
};

/******************************************************************************
 * Helpers
 *****************************************************************************/
static gboolean
xeme_message_localized_string_language_equal(gconstpointer a,
                                             gconstpointer b)
{
	const char *lang1 = NULL;
	const char *lang2 = NULL;

	if(XEME_IS_LOCALIZED_STRING((gpointer)a)) {
		XemeLocalizedString *l10nstr = (gpointer)a;

		lang1 = xeme_localized_string_get_language(l10nstr);
	}

	if(XEME_IS_LOCALIZED_STRING((gpointer)b)) {
		XemeLocalizedString *l10nstr = (gpointer)b;

		lang2 = xeme_localized_string_get_language(l10nstr);
	}

	return xeme_str_equal(lang1, lang2);
}

static void
xeme_message_marshal_language_collection(GListModel *collection,
                                         GString *str,
                                         const char *element)
{
	for(guint i = 0; i < g_list_model_get_n_items(collection); i++) {
		XemeLocalizedString *l10nstr = NULL;
		const char *lang = NULL;
		const char *content = NULL;

		l10nstr = g_list_model_get_item(collection, i);
		lang = xeme_localized_string_get_language(l10nstr);
		content = xeme_localized_string_get_string(l10nstr);

		if(xeme_str_is_empty(lang)) {
			g_string_append_printf(str, "<%s>%s</%s>", element, content,
			                       element);
		} else {
			g_string_append_printf(str, "<%s xml:lang=\"%s\">%s</%s>",
			                       element, lang, content, element);
		}

		g_clear_object(&l10nstr);
	}
}

/******************************************************************************
 * XemeStanza Implementation
 *****************************************************************************/
static char *
xeme_message_marshal(XemeStanza *stanza, G_GNUC_UNUSED GError **error) {
	XemeMessage *message = XEME_MESSAGE(stanza);
	GString *str = NULL;
	char *attrs = NULL;

	str = g_string_new("<message");

	attrs = xeme_stanza_marshal_attributes(stanza);
	if(!xeme_str_is_empty(attrs)) {
		g_string_append_printf(str, " %s>", attrs);
	} else {
		g_string_append(str, ">");
	}
	g_clear_pointer(&attrs, g_free);

	/* Now run through the collections. */
	xeme_message_marshal_language_collection(G_LIST_MODEL(message->bodies),
	                                         str, "body");
	xeme_message_marshal_language_collection(G_LIST_MODEL(message->subjects),
	                                         str, "subject");

	if(!xeme_str_is_empty(message->thread)) {
		if(!xeme_str_is_empty(message->thread_parent)) {
			g_string_append_printf(str, "<thread parent=\"%s\">%s</thread>",
			                       message->thread_parent, message->thread);
		} else {
			g_string_append_printf(str, "<thread>%s</thread>",
			                       message->thread);
		}
	}

	/* Finish up */
	g_string_append(str, "</message>");

	return g_string_free_and_steal(str);
}

/******************************************************************************
 * GObject Implementation
 *****************************************************************************/
G_DEFINE_FINAL_TYPE(XemeMessage, xeme_message, XEME_TYPE_STANZA)

static void
xeme_message_finalize(GObject *obj) {
	XemeMessage *message = XEME_MESSAGE(obj);

	g_clear_object(&message->bodies);
	g_clear_object(&message->subjects);

	g_clear_pointer(&message->thread, g_free);
	g_clear_pointer(&message->thread_parent, g_free);

	G_OBJECT_CLASS(xeme_message_parent_class)->finalize(obj);
}

static void
xeme_message_get_property(GObject *obj, guint param_id, GValue *value,
                          GParamSpec *pspec)
{
	XemeMessage *message = XEME_MESSAGE(obj);

	switch(param_id) {
	case PROP_BODIES:
		g_value_set_object(value, xeme_message_get_bodies(message));
		break;
	case PROP_MESSAGE_TYPE:
		g_value_set_enum(value, xeme_message_get_message_type(message));
		break;
	case PROP_SUBJECTS:
		g_value_set_object(value, xeme_message_get_subjects(message));
		break;
	case PROP_THREAD:
		g_value_set_string(value, xeme_message_get_thread(message));
		break;
	case PROP_THREAD_PARENT:
		g_value_set_string(value, xeme_message_get_thread_parent(message));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_message_set_property(GObject *obj, guint param_id, const GValue *value,
                          GParamSpec *pspec)
{
	XemeMessage *message = XEME_MESSAGE(obj);

	switch(param_id) {
	case PROP_MESSAGE_TYPE:
		xeme_message_set_message_type(message, g_value_get_enum(value));
		break;
	case PROP_THREAD:
		xeme_message_set_thread(message, g_value_get_string(value));
		break;
	case PROP_THREAD_PARENT:
		xeme_message_set_thread_parent(message, g_value_get_string(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_message_init(XemeMessage *message) {
	message->message_type = XEME_MESSAGE_TYPE_NORMAL;

	message->bodies = g_list_store_new(XEME_TYPE_LOCALIZED_STRING);
	message->subjects = g_list_store_new(XEME_TYPE_LOCALIZED_STRING);
}

static void
xeme_message_class_init(XemeMessageClass *klass) {
	GObjectClass *obj_class = G_OBJECT_CLASS(klass);
	XemeStanzaClass *stanza_class = XEME_STANZA_CLASS(klass);

	obj_class->finalize = xeme_message_finalize;
	obj_class->get_property = xeme_message_get_property;
	obj_class->set_property = xeme_message_set_property;

	stanza_class->marshal = xeme_message_marshal;

	/**
	 * XemeMessage:bodies:
	 *
	 * All of the bodies for this message.
	 *
	 * Since: 0.1
	 */
	properties[PROP_BODIES] = g_param_spec_object(
		"bodies", NULL, NULL,
		G_TYPE_LIST_MODEL,
		G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeMessage:message-type:
	 *
	 * The type of the message.
	 *
	 * Since: 0.1
	 */
	properties[PROP_MESSAGE_TYPE] = g_param_spec_enum(
		"message-type", NULL, NULL,
		XEME_TYPE_MESSAGE_TYPE,
		XEME_MESSAGE_TYPE_NORMAL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeMessage:subjects:
	 *
	 * All of the subjects for this message.
	 *
	 * Since: 0.1
	 */
	properties[PROP_SUBJECTS] = g_param_spec_object(
		"subjects", NULL, NULL,
		G_TYPE_LIST_MODEL,
		G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeMessage:thread:
	 *
	 * The thread identifier for this message.
	 *
	 * Since: 0.1
	 */
	properties[PROP_THREAD] = g_param_spec_string(
		"thread", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeMessage:thread-parent:
	 *
	 * The thread parent identifier for this message.
	 *
	 * Since: 0.1
	 */
	properties[PROP_THREAD_PARENT] = g_param_spec_string(
		"thread-parent", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	g_object_class_install_properties(obj_class, N_PROPERTIES, properties);
}

/******************************************************************************
 * Public API
 *****************************************************************************/
void
xeme_message_add_body(XemeMessage *message, const char *language,
                      const char *content)
{
	XemeLocalizedString *new_body = NULL;
	gboolean found = FALSE;
	guint position = 0;

	g_return_if_fail(XEME_IS_MESSAGE(message));
	g_return_if_fail(!xeme_str_is_empty(content));

	new_body = xeme_localized_string_new(language, content);
	found = g_list_store_find_with_equal_func(message->bodies, new_body,
	                                          xeme_message_localized_string_language_equal,
	                                          &position);

	/* We found an existing string for the language so we need to remove it. */
	if(found) {
		g_list_store_remove(message->bodies, position);
	}

	/* Add our new body. */
	g_list_store_append(message->bodies, new_body);

	/* The append adds a reference so we need to clear ours. */
	g_clear_object(&new_body);
}

void
xeme_message_add_subject(XemeMessage *message, const char *language,
                         const char *content)
{
	XemeLocalizedString *new_subject = NULL;
	gboolean found = FALSE;
	guint position = 0;

	g_return_if_fail(XEME_IS_MESSAGE(message));
	g_return_if_fail(!xeme_str_is_empty(content));

	new_subject = xeme_localized_string_new(language, content);
	found = g_list_store_find_with_equal_func(message->subjects, new_subject,
	                                          xeme_message_localized_string_language_equal,
	                                          &position);

	/* We found an existing string for the language so we need to remove it. */
	if(found) {
		g_list_store_remove(message->subjects, position);
	}

	/* Add our new subject. */
	g_list_store_append(message->subjects, new_subject);

	/* The append adds a reference so we need to clear ours. */
	g_clear_object(&new_subject);
}

GListModel *
xeme_message_get_bodies(XemeMessage *message) {
	g_return_val_if_fail(XEME_IS_MESSAGE(message), NULL);

	return G_LIST_MODEL(message->bodies);
}

XemeLocalizedString *
xeme_message_get_body(XemeMessage *message, const char *language) {
	GListModel *model = NULL;

	g_return_val_if_fail(XEME_IS_MESSAGE(message), NULL);

	model = G_LIST_MODEL(message->bodies);
	for(guint i = 0; i < g_list_model_get_n_items(model); i++) {
		XemeLocalizedString *l10nstr = NULL;
		const char *language1 = NULL;

		l10nstr = g_list_model_get_item(model, i);
		language1 = xeme_localized_string_get_language(l10nstr);
		if(xeme_str_equal(language1, language)) {
			return l10nstr;
		}

		g_clear_object(&l10nstr);
	}

	return NULL;
}

XemeLocalizedString *
xeme_message_get_subject(XemeMessage *message, const char *language) {
	GListModel *model = NULL;

	g_return_val_if_fail(XEME_IS_MESSAGE(message), NULL);

	model = G_LIST_MODEL(message->subjects);
	for(guint i = 0; i < g_list_model_get_n_items(model); i++) {
		XemeLocalizedString *l10nstr = NULL;
		const char *language1 = NULL;

		l10nstr = g_list_model_get_item(model, i);
		language1 = xeme_localized_string_get_language(l10nstr);
		if(xeme_str_equal(language1, language)) {
			return l10nstr;
		}

		g_clear_object(&l10nstr);
	}

	return NULL;
}

GListModel *
xeme_message_get_subjects(XemeMessage *message) {
	g_return_val_if_fail(XEME_IS_MESSAGE(message), NULL);

	return G_LIST_MODEL(message->subjects);
}

XemeMessageType
xeme_message_get_message_type(XemeMessage *message) {
	g_return_val_if_fail(XEME_IS_MESSAGE(message), XEME_MESSAGE_TYPE_UNKNOWN);

	return message->message_type;
}

void
xeme_message_set_message_type(XemeMessage *message, XemeMessageType type) {
	g_return_if_fail(XEME_IS_MESSAGE(message));

	if(message->message_type != type) {
		message->message_type = type;

		g_object_notify_by_pspec(G_OBJECT(message),
		                         properties[PROP_MESSAGE_TYPE]);
	}
}


const char *
xeme_message_get_thread(XemeMessage *message) {
	g_return_val_if_fail(XEME_IS_MESSAGE(message), NULL);

	return message->thread;
}

const char *
xeme_message_get_thread_parent(XemeMessage *message) {
	g_return_val_if_fail(XEME_IS_MESSAGE(message), NULL);

	return message->thread_parent;
}

XemeStanza *
xeme_message_new(void) {
	return g_object_new(XEME_TYPE_MESSAGE, NULL);
}

gboolean
xeme_message_remove_body(XemeMessage *message, const char *language) {
	XemeLocalizedString *needle = NULL;
	gboolean found = FALSE;
	guint position = 0;

	g_return_val_if_fail(XEME_IS_MESSAGE(message), FALSE);

	needle = xeme_localized_string_new(language, "needle");

	found = g_list_store_find_with_equal_func(message->bodies, needle,
	                                          xeme_message_localized_string_language_equal,
	                                          &position);

	if(found) {
		g_list_store_remove(message->bodies, position);
	}

	g_clear_object(&needle);

	return found;
}

gboolean
xeme_message_remove_subject(XemeMessage *message, const char *language) {
	XemeLocalizedString *needle = NULL;
	gboolean found = FALSE;
	guint position = 0;

	g_return_val_if_fail(XEME_IS_MESSAGE(message), FALSE);

	needle = xeme_localized_string_new(language, "needle");

	found = g_list_store_find_with_equal_func(message->subjects, needle,
	                                          xeme_message_localized_string_language_equal,
	                                          &position);

	if(found) {
		g_list_store_remove(message->subjects, position);
	}

	g_clear_object(&needle);

	return found;
}

void
xeme_message_set_thread(XemeMessage *message, const char *thread) {
	g_return_if_fail(XEME_IS_MESSAGE(message));

	if(g_set_str(&message->thread, thread)) {
		g_object_notify_by_pspec(G_OBJECT(message), properties[PROP_THREAD]);
	}
}

void
xeme_message_set_thread_parent(XemeMessage *message,
                               const char *thread_parent)
{
	g_return_if_fail(XEME_IS_MESSAGE(message));

	if(g_set_str(&message->thread_parent, thread_parent)) {
		g_object_notify_by_pspec(G_OBJECT(message),
		                         properties[PROP_THREAD_PARENT]);
	}
}
