/*
 * Copyright (C) 2023-2025 Xeme Developers
 *
 * Xeme is the legal property of its developers, whose names are too
 * numerous to list here. Please refer to the AUTHORS file distributed
 * with this source distribution
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, see
 * <https://www.gnu.org/licenses/>.
 */

#include "xemestanza.h"
#include "xemestring.h"

enum {
	PROP_0,
	PROP_CONTENT_TYPE,
	PROP_ID,
	PROP_TO,
	PROP_FROM,
	PROP_XML_LANG,
	N_PROPERTIES,
};
static GParamSpec *properties[N_PROPERTIES] = {NULL, };

typedef struct {
	char *content_type;
	char *id;
	char *to;
	char *from;
	char *xml_lang;
} XemeStanzaPrivate;

/******************************************************************************
 * GObject Implementation
 *****************************************************************************/
G_DEFINE_TYPE_WITH_PRIVATE(XemeStanza, xeme_stanza, G_TYPE_OBJECT)

static void
xeme_stanza_finalize(GObject *obj) {
	XemeStanza *stanza = XEME_STANZA(obj);
	XemeStanzaPrivate *priv = xeme_stanza_get_instance_private(stanza);

	g_clear_pointer(&priv->content_type, g_free);
	g_clear_pointer(&priv->from, g_free);
	g_clear_pointer(&priv->id, g_free);
	g_clear_pointer(&priv->to, g_free);
	g_clear_pointer(&priv->xml_lang, g_free);

	G_OBJECT_CLASS(xeme_stanza_parent_class)->finalize(obj);
}

static void
xeme_stanza_get_property(GObject *obj, guint param_id, GValue *value,
                         GParamSpec *pspec)
{
	XemeStanza *stanza = XEME_STANZA(obj);

	switch(param_id) {
	case PROP_CONTENT_TYPE:
		g_value_set_string(value, xeme_stanza_get_content_type(stanza));
		break;
	case PROP_FROM:
		g_value_set_string(value, xeme_stanza_get_from(stanza));
		break;
	case PROP_ID:
		g_value_set_string(value, xeme_stanza_get_id(stanza));
		break;
	case PROP_TO:
		g_value_set_string(value, xeme_stanza_get_to(stanza));
		break;
	case PROP_XML_LANG:
		g_value_set_string(value, xeme_stanza_get_xml_lang(stanza));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_stanza_set_property(GObject *obj, guint param_id, const GValue *value,
                         GParamSpec *pspec)
{
	XemeStanza *stanza = XEME_STANZA(obj);

	switch(param_id) {
	case PROP_CONTENT_TYPE:
		xeme_stanza_set_content_type(stanza, g_value_get_string(value));
		break;
	case PROP_FROM:
		xeme_stanza_set_from(stanza, g_value_get_string(value));
		break;
	case PROP_ID:
		xeme_stanza_set_id(stanza, g_value_get_string(value));
		break;
	case PROP_TO:
		xeme_stanza_set_to(stanza, g_value_get_string(value));
		break;
	case PROP_XML_LANG:
		xeme_stanza_set_xml_lang(stanza, g_value_get_string(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_stanza_init(G_GNUC_UNUSED XemeStanza *stanza) {
}

static void
xeme_stanza_class_init(XemeStanzaClass *klass) {
	GObjectClass *obj_class = G_OBJECT_CLASS(klass);

	obj_class->finalize = xeme_stanza_finalize;
	obj_class->get_property = xeme_stanza_get_property;
	obj_class->set_property = xeme_stanza_set_property;

	/**
	 * XemeStanza:content-type:
	 *
	 * The content type of this stanza.
	 *
	 * Since: 0.1
	 */
	properties[PROP_CONTENT_TYPE] = g_param_spec_string(
		"content-type", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeStanza:from:
	 *
	 * The sender of this stanza.
	 *
	 * Since: 0.1
	 */
	properties[PROP_FROM] = g_param_spec_string(
		"from", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeStanza:id:
	 *
	 * The identifier of this stanza.
	 *
	 * Since: 0.1
	 */
	properties[PROP_ID] = g_param_spec_string(
		"id", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeStanza:to:
	 *
	 * The recipient of this stanza.
	 *
	 * Since: 0.1
	 */
	properties[PROP_TO] = g_param_spec_string(
		"to", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeStanza:xml-lang:
	 *
	 * The XML Lang value for this stanza.
	 *
	 * Since: 0.1
	 */
	properties[PROP_XML_LANG] = g_param_spec_string(
		"xml-lang", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	g_object_class_install_properties(obj_class, N_PROPERTIES, properties);
}

/******************************************************************************
 * Public API
 *****************************************************************************/
const char *
xeme_stanza_get_content_type(XemeStanza *stanza) {
	XemeStanzaPrivate *priv = NULL;

	g_return_val_if_fail(XEME_IS_STANZA(stanza), NULL);

	priv = xeme_stanza_get_instance_private(stanza);

	return priv->content_type;
}

void
xeme_stanza_set_content_type(XemeStanza *stanza, const char *content_type) {
	XemeStanzaPrivate *priv = NULL;

	g_return_if_fail(XEME_IS_STANZA(stanza));
	g_return_if_fail(content_type != NULL);

	priv = xeme_stanza_get_instance_private(stanza);

	if(g_set_str(&priv->content_type, content_type)) {
		g_object_notify_by_pspec(G_OBJECT(stanza),
		                         properties[PROP_CONTENT_TYPE]);
	}
}

const char *
xeme_stanza_get_from(XemeStanza *stanza) {
	XemeStanzaPrivate *priv = NULL;

	g_return_val_if_fail(XEME_IS_STANZA(stanza), NULL);

	priv = xeme_stanza_get_instance_private(stanza);

	return priv->from;
}

void
xeme_stanza_set_from(XemeStanza *stanza, const char *from) {
	XemeStanzaPrivate *priv = NULL;

	g_return_if_fail(XEME_IS_STANZA(stanza));
	g_return_if_fail(from != NULL);

	priv = xeme_stanza_get_instance_private(stanza);

	if(g_set_str(&priv->from, from)) {
		g_object_notify_by_pspec(G_OBJECT(stanza), properties[PROP_FROM]);
	}
}

const char *
xeme_stanza_get_id(XemeStanza *stanza) {
	XemeStanzaPrivate *priv = NULL;

	g_return_val_if_fail(XEME_IS_STANZA(stanza), NULL);

	priv = xeme_stanza_get_instance_private(stanza);

	return priv->id;
}

void
xeme_stanza_set_id(XemeStanza *stanza, const char *id) {
	XemeStanzaPrivate *priv = NULL;

	g_return_if_fail(XEME_IS_STANZA(stanza));
	g_return_if_fail(id != NULL);

	priv = xeme_stanza_get_instance_private(stanza);

	if(g_set_str(&priv->id, id)) {
		g_object_notify_by_pspec(G_OBJECT(stanza), properties[PROP_ID]);
	}
}

const char *
xeme_stanza_get_to(XemeStanza *stanza) {
	XemeStanzaPrivate *priv = NULL;

	g_return_val_if_fail(XEME_IS_STANZA(stanza), NULL);

	priv = xeme_stanza_get_instance_private(stanza);

	return priv->to;
}

void
xeme_stanza_set_to(XemeStanza *stanza, const char *to) {
	XemeStanzaPrivate *priv = NULL;

	g_return_if_fail(XEME_IS_STANZA(stanza));
	g_return_if_fail(to != NULL);

	priv = xeme_stanza_get_instance_private(stanza);

	if(g_set_str(&priv->to, to)) {
		g_object_notify_by_pspec(G_OBJECT(stanza),
		                         properties[PROP_TO]);
	}
}

const char *
xeme_stanza_get_xml_lang(XemeStanza *stanza) {
	XemeStanzaPrivate *priv = NULL;

	g_return_val_if_fail(XEME_IS_STANZA(stanza), NULL);

	priv = xeme_stanza_get_instance_private(stanza);

	return priv->xml_lang;
}

void
xeme_stanza_set_xml_lang(XemeStanza *stanza, const char *xml_lang) {
	XemeStanzaPrivate *priv = NULL;

	g_return_if_fail(XEME_IS_STANZA(stanza));

	priv = xeme_stanza_get_instance_private(stanza);

	if(g_set_str(&priv->xml_lang, xml_lang)) {
		g_object_notify_by_pspec(G_OBJECT(stanza), properties[PROP_XML_LANG]);
	}
}

char *
xeme_stanza_marshal(XemeStanza *stanza, GError **error) {
	XemeStanzaClass *klass = NULL;

	g_return_val_if_fail(XEME_IS_STANZA(stanza), NULL);
	g_return_val_if_fail(error != NULL && *error == NULL, NULL);

	klass = XEME_STANZA_GET_CLASS(stanza);
	if(klass != NULL && klass->marshal != NULL) {
		return klass->marshal(stanza, error);
	}

	return NULL;
}

char *
xeme_stanza_marshal_attributes(XemeStanza *stanza) {
	XemeStanzaPrivate *priv = NULL;
	GString *str = NULL;

	g_return_val_if_fail(XEME_IS_STANZA(stanza), NULL);

	priv = xeme_stanza_get_instance_private(stanza);
	str = g_string_new("");

	/* Attributes should remain in alphabetical order for consistency. */

	if(!xeme_str_is_empty(priv->from)) {
		g_string_append_printf(str, " from=\"%s\"", priv->from);
	}

	if(!xeme_str_is_empty(priv->id)) {
		g_string_append_printf(str, " id=\"%s\"", priv->id);
	}

	if(!xeme_str_is_empty(priv->to)) {
		g_string_append_printf(str, " to=\"%s\"", priv->to);
	}

	if(!xeme_str_is_empty(priv->content_type)) {
		g_string_append_printf(str, " type=\"%s\"", priv->content_type);
	}

	if(!xeme_str_is_empty(priv->xml_lang)) {
		g_string_append_printf(str, " xml:lang=\"%s\"", priv->xml_lang);
	}

	/* Finally remove the leading space that we added to make appending
	 * everything easier.
	 */
	if(str->len > 0) {
		g_string_erase(str, 0, 1);
	}

	return g_string_free_and_steal(str);
}
